/*
 * pfcoll: pf log event collector
 *
 * Copyright (c) Carson Harding, 2003-2006.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are freely permitted.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
 * THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <err.h>
#include <poll.h>
#include <pwd.h>

const char *rcsid = "$Id: pfcoll.c,v 1.20 2014/05/16 17:16:02 harding Exp $";

#define _PFCOLL_STATS	"/var/db/pfcoll.stats"

#define _PFCOLL_FAST	1	/* connections/sec to be high rate */
#define _PFCOLL_TOOFAST	15	/* fast connections for more than secs */
#define _PFCOLL_RECOVER	60	/* secs before consider rate has recovered */

/*
 * - Check for pending messages every 10 seconds (in absence of intervening
 *   bump events).
 * - If bumps have been accumulating for 10 or more seconds, log them.
 * - If bumps have been accumulating for 1 minute or more, send them.
 */
#define _PFCOLL_IO_TIMEO	10
#define _PFCOLL_MESG_TIMEO	10	   /* log pending until seconds */
#define _PFCOLL_MAIL_TIMEO	(1*60)	   /* send pending until seconds */

#define _PFCOLL_MESG_ONCE	1
#define _PFCOLL_MESG_SINCELAST	2

#define _PFCOLL_SENDMAIL	"/usr/sbin/sendmail"
#define _PFCOLL_MAIL_ONCE	1
#define _PFCOLL_MAIL_SINCELAST	2

#define _PFCOLL_TCPDUMP		"/usr/sbin/tcpdump"
#define _PFCOLL_LINESZ		2048	/* size of line output from tcpdump */
#define _PFCOLL_MAXLINES	32	/* max lines to mail */
#define _PFCOLL_NOLINES		0	/* don't output collected lines */
#define _PFCOLL_DOLINES		1	/* do output collected lines */

struct llist {
	char *line;
	struct llist *next;
};

typedef struct connection_stats connstats;
struct	connection_stats {
	int	forks;		/* children spawned */
	int	total;		/* total # of connections */
	time_t	atstart;	/* start time  */
	time_t	now;		/* time of this connection */
	time_t	last;		/* time of last connection */
	time_t	lastmesg;	/* last mesg we logged */
	time_t	lastmail;	/* last mail we sent */
	int	sincemail;	/* connections since last mail */
	int	sincemesg;	/* connections since last mesg */
	int	fast;		/* how many at a fast rate? */
	int	toofast;	/* has it been too fast? */ 
	struct	llist *lhead;	/* linked list of log lines */
	struct	llist *ltail;
	int	lcount;		/* current number of lines in list */
	char	*line;
};

int	f_debug;		/* debugging: no daemon, log to stderr */
int	f_forking = 1;		/* fork to process connection */
int	f_daemon = 1;		/* run as daemon */
char	*pidfile;		/* write to this pid file */
pid_t	parent;			/* all logging references master */
pid_t	tdumppid;		/* tcpdump pid */
char	*logfile;		/* logfile to use */
char	*mailaddr;		/* who to mail for reports */
char	*mesgstr;		/* short message to identify source */
int	newac;			/* argc, argv for exec */
char	**newav;
#define START_AV_SZ	16

volatile sig_atomic_t	wantreap;	/* child exited, need to reap */
volatile sig_atomic_t 	wantstats;	/* request for stats received */

extern	char *__progname;
char	*progversion = VER;
char	*proguser;

void	usage(void);
void	add_arg(char *s);
FILE 	*start_tcpdump(char **av);
void	handle_line(connstats *cn, char *line);
void	pending_messages(connstats *cn);
void	handle_messages(connstats *cn, int mesgtype, int mailtype); 
char	*formatted_time(time_t *then);
void	write_pid(char *pidfile);
void	remove_pid(void);
void	mail(int mailtype, char *address, connstats *cn);
void	sigchld(int sig);
void	sigusr1(int sig);
void	sigalrm(int sig);
void	sigexit(int sig);
void	reap(void);
void	addllist(connstats *cn, char *line);
void	freellist(connstats *cn);
void	log_stats(connstats *cn);
void	print_stats(connstats *cn, FILE *fw, int dologlines);
void	errlog(int level, char *fmt, ...);
void	doerrlog(int level, char *fmt, va_list ap);
void	log_vprintf(FILE *fw, char *fmt, va_list ap);

void
usage(void)
{
	fprintf(stderr, 
	    "usage: %s [-V] [-DF] [-m email] [-M mesg] [-P pidfile] [-l log] "
#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	    " [-u user] "
#endif
	    "[-i if] {-s|expression...}\n", 
	    __progname);
	fprintf(stderr, "       -V print version and exit\n");
	fprintf(stderr, "       -D debug, stay in foreground\n");
	fprintf(stderr, "       -F no fork on event\n");
	fprintf(stderr, "       -M short message to put in mail\n");
	fprintf(stderr, "       -P write pid to this file\n");
	fprintf(stderr, "       -m mail reports to\n");
	fprintf(stderr, "       -l log (instead of syslog)\n");
#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	fprintf(stderr, "       -u user to become\n");
#endif
	fprintf(stderr, "       -i interface (default pflog0)\n");
	fprintf(stderr, "       -s read from stdin, not tcpdump\n");
	
	exit(1);
}

void
version(void)
{
	fprintf(stderr, "%s version %s\n", __progname, progversion);
	exit(0);
}

int
main(int argc, char **argv)
{
	int	i;
	int	ch;
	char	*logif = "pflog0";
	char	line[_PFCOLL_LINESZ];
	struct	sigaction act;
	struct	passwd *pwent;
	connstats conn;
	FILE	*pr = NULL;
#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	char	*user = NULL;
#endif

	bzero(&conn, sizeof(conn));

	/* process options */
#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	while ((ch = getopt(argc, argv, "DFM:P:Vi:l:m:su:")) != -1) {
#else
	while ((ch = getopt(argc, argv, "DFM:P:Vi:l:m:s")) != -1) {
#endif
		switch(ch) {
		case 'D':
			f_debug = 1;
			f_daemon = 0;
			break;
		case 'F':
			f_forking = 0;
			break;
		case 'M':
			mesgstr = strdup(optarg);
			if (!mesgstr)
				err(1, "cannot allocate for message string");
			break;
		case 'P':
			pidfile = strdup(optarg);
			if (!pidfile)
				err(1, "cannot allocate for pid file path");
			break;
		case 'V':
			version();
			break;
		case 'i':
			logif = optarg;
			break;
		case 'l':
			logfile = optarg;
			break;
		case 'm':
			mailaddr = optarg;
			break;
#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
		case 'u':
			user = optarg;
			break;
#endif
		case 's':
			/* input testing */
			pr = stdin;
			f_daemon = 0;
			break;
		default:
			usage();
			break;
		}
	}

	argc -= optind;
	argv += optind;

	pwent = getpwuid(getuid());
	if (!pwent)
		proguser = "unknown";
	else
		proguser = pwent->pw_name;

#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	if (user) {
		uid_t uid;
		if (getuid() != 0)
			warnx("cannot switch users if not root!");
		pwent = getpwnam(user);
		if (!pwent && isdigit(*user)) {
			uid = (uid_t)strtol(user, (char **)NULL, 10);
			pwent = getpwuid(uid);
		}
		if (!pwent)
			err(1, "%s", user);
	}
#endif

	if (f_daemon) {
		if (daemon(0, 0) == -1)
			err(1, "cannot become daemon");
		openlog(__progname, LOG_PID, LOG_USER);
		if (pidfile)
			write_pid(pidfile);
	}

	if (!pr) {
		add_arg("tcpdump");
		add_arg("-en");
		add_arg("-l");
		add_arg("-i");
		add_arg(logif);
#if defined(TCPDUMP_SWITCH_USER)
		if (user) {
			add_arg("-u");
			add_arg(user);
		}
#endif	
		for (i = 0; i < argc; i++)
			add_arg(argv[i]);
		pr = start_tcpdump(newav);
		if (!pr)
			err(1, "tcpdump");
	}

#if defined(TCPDUMP_SWITCH_USER) || defined(PFCOLL_SWITCH_USER)
	if (user) {
		setgroups(1, &pwent->pw_gid);
		setegid(pwent->pw_gid);
		setgid(pwent->pw_gid);
		seteuid(pwent->pw_uid);
		setuid(pwent->pw_uid);
	}
#endif

	parent = getpid();

	(void)time(&conn.atstart);
	errlog(LOG_INFO, "started; tcpdump pid = %d", (int)tdumppid);

	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	act.sa_handler = sigchld;
	sigaction(SIGCHLD, &act, NULL);
	act.sa_handler = sigusr1;
	sigaction(SIGUSR1, &act, NULL);
	act.sa_handler = sigalrm;
	sigaction(SIGALRM, &act, NULL);
	act.sa_handler = sigexit;
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGINT, &act, NULL);

	/* kind-of silly */
	conn.line = line;

	for (;;) {
		if (wantreap)
			reap();
		if (wantstats)
			log_stats(&conn);
		alarm(_PFCOLL_IO_TIMEO);
		fgets(line, sizeof(line), pr);
		if (ferror(pr)) {
			switch(errno) {
			case EINTR:
				pending_messages(&conn);
				clearerr(pr);
				break;
			case EAGAIN:
				break;
			default:
				errlog(LOG_ERR, "error on tcpdump pipe: %s",
				    strerror(errno));
				exit(1);
			}
		} else if (feof(pr)) {
			errlog(LOG_INFO, "end of input, exiting");
			break;
		}
		else
			handle_line(&conn, line);
	}

	exit(0);
	return 0;
}

/*
 * Add an argument to the argument array.
 */
void
add_arg(char *s) 
{
	char	*p;
	size_t	len;
	static	size_t newamax = START_AV_SZ;

	len = strlen(s);

	if (!newav) {
		newav = calloc(START_AV_SZ, sizeof(char *));
		if (!newav)
			err(1, "malloc");
	} else if (newac >= newamax-1) {
		newamax *= 2;
		newav = realloc(newav, newamax * sizeof(char *));
		if (!newav)
			err(1, "realloc");
	}
	p = malloc(len+1);		
	if (!p) err(1, "malloc");
	memmove(p, s, len);
	p[len] = '\0';
	newav[newac++] = p;
	newav[newac] = NULL;
	
	return;
}

void
handle_line(connstats *cn, char *line)
{
	int	mesgtype = 0;
	int	mailtype = 0;
	double	last;

	errlog(LOG_DEBUG, "handle_line(): entering");

	cn->total++;
	time(&cn->now);
	last = difftime(cn->now, cn->last);
	if (last < _PFCOLL_FAST)
		cn->fast++;
	else if (last > _PFCOLL_RECOVER) {
		cn->fast = 0;
		cn->toofast = 0;
	} 
	if (cn->fast > _PFCOLL_TOOFAST)
		cn->toofast = 1;
	cn->last = cn->now;

	/* if too many log entries, don't add more to list */
	if (cn->lcount <= _PFCOLL_MAXLINES)
		addllist(cn, line);

	/* log messages */
	if (difftime(cn->now, cn->lastmesg) < _PFCOLL_MESG_TIMEO) {
		/* too soon */
		cn->sincemesg++;
		errlog(LOG_DEBUG, 
		    "handle_line(): queuing log message");
	} else if (!cn->sincemesg) {
		mesgtype = _PFCOLL_MESG_ONCE;
		errlog(LOG_DEBUG, 
		    "handle_line(): setting "
		    "mesgtype to handle this log message");
	} else {
		mesgtype = _PFCOLL_MESG_SINCELAST;
		errlog(LOG_DEBUG, 
		    "handle_line(): setting "
		    "mesgtype to handle this and past log messages");
	}

	/* mail messages */
	if (mailaddr) {
		if (difftime(cn->now, cn->lastmail) < _PFCOLL_MAIL_TIMEO) {
			/* too soon */
			cn->sincemail++;
			errlog(LOG_DEBUG, 
			    "handle_line(): queuing for mail");
		} else if (!cn->sincemail) {
			mailtype = _PFCOLL_MAIL_ONCE;
			errlog(LOG_DEBUG, 
			    "handle_line(): setting "
			    "mailtype to handle this message");
		} else {
			mailtype = _PFCOLL_MAIL_SINCELAST;
			errlog(LOG_DEBUG, 
			    "handle_line(): setting "
			    "mailtype to handle this and past messages");
		}
	}

	handle_messages(cn, mesgtype, mailtype);
}

/*
 * If there have been connections since the last message or mail
 * time, and time has expired, post the messages and/or mail.
 */
void
pending_messages(connstats *cn)
{
	int	mesgtype = 0;
	int	mailtype = 0;

	errlog(LOG_DEBUG, "pending_messages(): entering");

	(void)time(&cn->now);

	/* log messages */
	if (difftime(cn->now, cn->lastmesg) >= _PFCOLL_MESG_TIMEO
	    && cn->sincemesg) {
		mesgtype = _PFCOLL_MESG_SINCELAST;
		errlog(LOG_DEBUG, 
		    "pending_messages(): setting "
		    "mesgtype to handle pending log messages");
	}

	/* mail messages */
	if (mailaddr
	    && difftime(cn->now, cn->lastmail) >= _PFCOLL_MAIL_TIMEO
	    && cn->sincemail) {
		mailtype = _PFCOLL_MAIL_SINCELAST;
		errlog(LOG_DEBUG, 
		    "pending_messages(): setting "
		    "mailtype to handle pending messages");
	}

	handle_messages(cn, mesgtype, mailtype);
}

void
handle_messages(connstats *cn, int mesgtype, int mailtype)
{

	errlog(LOG_DEBUG, "handle_messages(): entering");

	/* don't fork if we have nothing to do! */
	if (!mesgtype && !mailtype)
		return;

	if (mesgtype) {
		switch(mesgtype) {
		case _PFCOLL_MESG_ONCE:
			errlog(LOG_INFO, 
			    "bump: %s", cn->line); 
			break;
		case _PFCOLL_MESG_SINCELAST:
			errlog(LOG_INFO, 
			    "bump %d times since last mesg; "
			    "last at %s", 
			    cn->sincemesg,
			    formatted_time(&cn->last));
			errlog(LOG_INFO, 
			    "this last bump: %s", cn->line); 
			break;
		default:
			errlog(LOG_ERR, "BUG! bad message type %d\n", 
			    mesgtype);
			break;
		}
		cn->sincemesg = 0;
		cn->lastmesg = cn->now;
	}

	if (mailaddr && mailtype) {
		if (f_forking) {
			switch(fork()) {
			case 0:
				break;
			case -1:
				errlog(LOG_ERR, "fork: %s", strerror(errno));
				/* we fail to handle new connection */	 
				return;
			default:
				/* parent */
				if (mailtype) {
					freellist(cn);
					cn->sincemail = 0;
					cn->lastmail = cn->now;
				}
				cn->forks++;
				return;
			}
			/* child : exit if around too long */
			signal(SIGALRM, SIG_DFL);
			alarm(30);
		}

		mail(mailtype, mailaddr, cn);

		if (f_forking)
			_exit(0);	
	}
}

/*
 * Nicely formatted time string for logging
 */
char *
formatted_time(time_t *then)
{
	static	char timestr[32];
	time_t  now;
	struct	tm *tm;

	if (then)
		now = *then;
	else {
		(void)time(&now);
	}
	tm = localtime(&now);
		
	(void)strftime(timestr, sizeof(timestr), 
	    "%Y/%m/%d %H:%M:%S", tm);

	return timestr;
}

void
write_pid(char *pidfile)
{
	FILE	*pidf;

	if ((pidf = fopen(pidfile, "w")) != NULL) {
		(void)atexit(remove_pid);
		fprintf(pidf, "%d\n", (int)getpid());
		(void)fclose(pidf);
	}
}

void
remove_pid(void)
{
	(void)unlink(pidfile);
}

FILE *
start_tcpdump(char *av[])
{
	char	**sv;
	FILE *pr = NULL;
	int tstderr;
	int tpdes;
	int pdes[2];

	errlog(LOG_DEBUG, "starting tcpdump");

	if (f_debug) {
		for (sv = av; *sv != NULL; sv++)
			fprintf(stderr, "%s ", *sv);
		fprintf(stderr, "\n");
	}

	if (pipe(pdes) != 0) {
		errlog(LOG_ERR, "pipe: %s", strerror(errno));
		return NULL;
	}

	switch(tdumppid = vfork()) {
	case -1:
		close(pdes[0]);
		close(pdes[1]);
		errlog(LOG_ERR, "fork: %s", strerror(errno));
		return NULL;
	case 0:
		tpdes = pdes[1];
		close(pdes[0]);
		if (tpdes != STDOUT_FILENO) {
			dup2(tpdes, STDOUT_FILENO);
			close(tpdes);
		}
		tstderr = open("/dev/null", O_WRONLY, 0);
		if (tstderr >= 0) {
			dup2(tstderr, STDERR_FILENO);
			close(tstderr);
		}
		if (execv(_PFCOLL_TCPDUMP, av) < 0) {
			errlog(LOG_ERR, 
			    "cannot exec " _PFCOLL_TCPDUMP ": %s", 
			    strerror(errno));
		}
		_exit(127);
	}

	/* parent */
	(void)close(pdes[1]);
	pr = fdopen(pdes[0], "r");
	if (pr == NULL) {
		errlog(LOG_ERR, "fdopen: %s", strerror(errno));
		return NULL;
	}

	return pr;
}

void
mail(int mailtype, char *address, connstats *cn)
{
	char *argv[4]= {"sendmail", "-oi", "-t", NULL};
	FILE *pw = NULL;
	int pdes[2];
	pid_t pid;

	errlog(LOG_DEBUG, "sending mail to %s", address);

	if (pipe(pdes) != 0) {
		errlog(LOG_ERR, "pipe: %s", strerror(errno));
		return;
	}

	switch(pid = fork()) {
	case -1:
		errlog(LOG_ERR, "fork: %s", strerror(errno));
		return;
	case 0:
		/* child */
		(void)close(pdes[1]);
		if (pdes[0] != STDIN_FILENO) {
			(void)dup2(pdes[0], STDIN_FILENO);
			(void)close(pdes[0]);
		}
		if (execv(_PFCOLL_SENDMAIL, argv) < 0) {
			errlog(LOG_ERR, 
			    "cannot exec " _PFCOLL_SENDMAIL ": %s",
			    strerror(errno));
		}
		_exit(1);
	}

	/* parent */
	(void)close(pdes[0]);
	pw = fdopen(pdes[1], "w");
	if (pw == NULL) {
		errlog(LOG_ERR, "fdopen: %s", strerror(errno));
		return;
	}

	fprintf(pw, "To: %s\n", address);
	fprintf(pw, "From: %s\n", proguser);
	fprintf(pw, "Subject: ** bump! at %s\n", formatted_time(&cn->last));
	fprintf(pw, "Organization: Department of "
	    "things that go bump in the night\n");
	fprintf(pw, "X-Mailer: %s %s\n", __progname, progversion);
	fprintf(pw, "MIME-Version: 1.0\n");
	fprintf(pw, "Content-Type: text/plain\n");
	fprintf(pw, "\n");

	if (mesgstr)
		fprintf(pw, "**  %s **\n\n", mesgstr);

	switch(mailtype) {
	case _PFCOLL_MAIL_ONCE:
		fprintf(pw, "There was a bump at %s\n",
		    formatted_time(&cn->last));
		fprintf(pw, "\n");
		break;
	case _PFCOLL_MAIL_SINCELAST:
		fprintf(pw, "There have been %d bumps since last mail.\n",
		    cn->sincemail);
		fprintf(pw, "\n");
		fprintf(pw, "The last bump was at %s\n",
		    formatted_time(&cn->last));
		fprintf(pw, "\n");
		break;
	default:
		fprintf(pw, "BUG: bad mail type!\n");
		errlog(LOG_ERR, "BUG: bad mail type!");
		break;
	}

	print_stats(cn, pw, _PFCOLL_DOLINES);

	fclose(pw);
	close(pdes[1]);
}

void
sigusr1(int sig)
{
	wantstats = 1;
}

void
sigchld(int sig)
{
	wantreap = 1;
}

void
sigalrm(int sig)
{
	return;
}

void
sigexit(int sig)
{
	/*
	 * XXX does not cooperate well with -u as written without
	 * XXX a user-switching tcpdump.
	 */
	kill(tdumppid, SIGTERM);
	_exit(0);
}

void
reap(void)
{	
	int	status;

	/* (void)wait3(&status, WNOHANG, NULL); */
	(void)wait3(&status, 0, NULL);	
	wantreap = 0;	
}

/* 
 * add line to list
 */
void
addllist(connstats *cn, char *line)
{
	struct	llist *ll;

	ll = malloc(sizeof(struct llist));
	if (!ll) {
		errlog(LOG_ERR, "malloc: %s", strerror(errno));
		exit(1);
	}
	ll->line = strdup(line);
	if (!ll->line) {
		errlog(LOG_ERR, "malloc: %s", strerror(errno));
		exit(1);
	}
	ll->next = NULL;
	if (cn->ltail)
		cn->ltail->next = ll;
	cn->ltail = ll;
	if (!cn->lhead)
		cn->lhead = ll;
	cn->lcount++;
}

void
freellist(connstats *cn)
{
	struct llist *lp, *lq;

	for (lp = cn->lhead; lp; lp=lq) {
		lq = lp->next;
		free(lp->line);
		free(lp);
	}
	cn->lhead = NULL;
	cn->ltail = NULL;
	cn->lcount = 0;
}

void
log_stats(connstats *cn)
{
	FILE	*logf;
	char	*log;

	wantstats = 0;

	if (f_debug) {
		/* show lines when debugging */
		print_stats(cn, stderr, _PFCOLL_DOLINES);
	} else {
		if (logfile)
			log = logfile;
		else
			log = _PFCOLL_STATS;
		if ((logf = fopen(log, "a")) != NULL) {
			print_stats(cn, logf, _PFCOLL_NOLINES);
			fputs("--------------------------------"
			      "--------------------------------"
                              "\n", logf);
			fclose(logf);
		}
	}
}

void
print_stats(connstats *cn, FILE *fw, int dologlines)
{
	struct llist *ll;

	if (dologlines) {
		fprintf(fw, "%d log line%s follow%s", 
		    cn->lcount, 
		    (cn->lcount == 1) ? "" : "s",
		    (cn->lcount == 1) ? "s" : "");
		if (cn->lcount > _PFCOLL_MAXLINES) {
			fprintf(fw, " (too many: %d events not saved)",
				cn->sincemail - cn->lcount);
		}
		fprintf(fw, ":\n\n");
		for (ll = cn->lhead; ll; ll=ll->next) {
			fprintf(fw, "%s", ll->line);
		}
	}

	fprintf(fw, "\n");
	fprintf(fw, "Stats for %s[%d] at %s:\n", 
	     __progname, (int)parent, formatted_time(&cn->now));
	fprintf(fw, "    start:        %s\n", formatted_time(&cn->atstart));
	fprintf(fw, "    lastbump:     %s\n",
	    cn->last ? formatted_time(&cn->last) : "never");
	fprintf(fw, "    total:        %d\n", cn->total);
	fprintf(fw, "    fast:         %d\n", cn->fast);
	fprintf(fw, "    toofast:      %d\n", cn->toofast);
	fprintf(fw, "    forks:        %d\n", cn->forks);
	fprintf(fw, "    lastlog:      %s\n", 
	    cn->lastmesg ? formatted_time(&cn->lastmesg) : "never");
	fprintf(fw, "    sincelog:     %d\n", cn->sincemesg);
	fprintf(fw, "    lastmail:     %s\n",
	    cn->lastmail ? formatted_time(&cn->lastmail) : "never");
	fprintf(fw, "    sincemail:    %d\n", cn->sincemail);

}

void
errlog(int level, char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	doerrlog(level, fmt, ap);
	va_end(ap);
}

void
doerrlog(int level, char *fmt, va_list ap)
{
	FILE	*logf;

	if (f_debug) {
		log_vprintf(stderr, fmt, ap);
	} else if (logfile) {
		if ((logf = fopen(logfile, "a")) != NULL) {
			log_vprintf(logf, fmt, ap);
			fclose(logf);
		}
	} else {
		vsyslog(level, fmt, ap);
	}
}

void
log_vprintf(FILE *fw, char *fmt, va_list ap)
{
	fprintf(fw, "%s: %s[%d]: ",
	    formatted_time(NULL), __progname, (int)parent); 
	vfprintf(fw, fmt, ap);
	fprintf(fw, "\n");
}


/* END */
